"""PHP-style authorisation functions.

Version:

    File Version 0.1 for LEMON 0.2
    
Usage:

    session = lemon.session.SessionDB(cursor, appName)  # Retrieve the session or create a new one
    auth    = lemon.auth.AuthDB(session)                # Create auth object from the session information
    
    if auth.isValid():                                  # See if the User is logged on else present login form.
        user = a.getUser()                              # Return a user object containing user's details.
        user.Password                                   # Access the data via attributes or dict style.
        user.getStatus()                                # Retrieve the entry for the appName column.

ToDO:

    addUser, RemoveUser etc.
    gadfly
    create the database if not already there?? 
    definately need config file similar to php.ini
    
    Sort out the ridiculous template situation with emailing etc.
"""
__docformat__ = "plaintext en"
__author__ = "James Gardner [email: james at xecos.com]"
__date__ = "02 July 2003"
__version__ = "Version 0.1 for LEMON 0.2"[-3:]
__credits__ = "Guido van Rossum, for an excellent programming language."

import lemon, lemon.session, time, os, cgi, lemon.session
import drivers.db
import template
                    
class AuthError(Exception):
    """Error Class for the Session Module. Use as follows:

    try:
        raise AuthError(ERROR_PASSWORD)
    except SessionError, e:
        print 'Auth exception occurred, value:', e.value
    """
    
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return self.value

def start(
    session, 
    storage='db',
    database=None,
    expire=14400, 
    idle=6000, 
    loginForm=None,
    autoLogin=True,
    autoRedirect=True,
    redirect=None,
    includeQuery=False,
    stickyData={},
    reminderForm=None,
    email=None,
    app=None,
    accessDenied=None,
    emailMessage=None,
    htmlPage=None,
):
    "The preffered method to start the auth classes."
    
    if storage == 'db':
        if not database:
            raise AuthError("No database connection specified for 'db' driver.")
        driver = drivers.db.DB(database)
    else:
        raise AuthError("You must specify a driver. 'db' is the only driver currently available.")
    
    return Auth( 
        session, 
        driver, 
        expire, 
        idle, 
        loginForm, 
        autoLogin, 
        autoRedirect, 
        redirect, 
        includeQuery, 
        stickyData, 
        reminderForm, 
        email, 
        app, 
        accessDenied, 
        emailMessage, 
        htmlPage
    )

class Auth:
    def __init__(
        self, 
        session, 
        driver, 
        expire, 
        idle, 
        loginForm, 
        autoLogin, 
        autoRedirect=True,
        redirect=None, 
        includeQuery=False, 
        stickyData={}, 
        reminderForm=None, 
        email=None, 
        app=None, 
        accessDenied=None, 
        emailMessage=None, 
        htmlPage=None
    ):
        """
        Parameters:
            session     - Session from lemon.session.start()
            driver      - Driver class
            expire      - Maximum length of time the user is allowed to be logged in for
            idle        - Maximum length of time between user requests
            loginForm   - Function to display get the HTML for the login form
            autoLogin   - Whether or not to automatically login the user
            redirect    - Where to go when logged in
            email = {
                'system':config.get('general','systemName'),
                'sender':config.get('admin','sender'),
                'reply':config.get('admin','reply'),
                'sendmail':config.get('admin','sendmail'),
                'smtp':config.get('admin','smtp'),
                'method':config.get('admin','method'),
            }
        """
        self.user = {'username':'', 'password':'','level':''}
        self.session = session
        self.app = app
        if not app:
            self.app = self.session.getParam('name')
        self.driver = driver
        
        self.expire = int(expire)
        self.idle = int(idle)
        self.stickyData = stickyData
        self.email = email
        
       
        if accessDenied:
            self.accessDenied = accessDenied
        else:
            self.accessDenied = template.accessDenied
            
        if loginForm:
            self.loginForm = loginForm
        else:
            self.loginForm = template.loginForm
            
        if reminderForm:
            self.reminderForm = reminderForm
        else:
            self.reminderForm = template.reminderForm
            
        if emailMessage:
            self.emailMessage = emailMessage
        else:
            self.emailMessage = template.emailMessage
            
        if htmlPage:
            self.htmlPage = htmlPage
        else:
            self.htmlPage = template.htmlPage
            
        self.autoLogin = autoLogin
        

        self.redirect = redirect
        self.autoRedirect = autoRedirect
        self.includeQuery = includeQuery
            

        # Block App 'Auth' so it can't be accidently over-written.
        session.blockVariable('AuthRedirectURL')
        session.blockVariable('AuthStarted')
        session.blockVariable('AuthAccessed')
        session.blockVariable('AuthIdleLength')
        session.blockVariable('AuthExpireLength')
        session.blockVariable('AuthUser')


        self.errorCodes = {      0: ('ACCESS_DENIED', 'You do not have access rights to use this application.'),
                                -1: ('EXPIRED','The session expired and you have been logged out.'),
                                -2: ('IDLED', 'The session has been left idle for too long and you have been logged out.',),
                                -3: ('NO_USER','The username specified in your session does not exist. Try signing in again.'),
                                -4: ('NO_SESSION','There is no session information for the username you are using. This may be because you have been logged out.'),
                                -5: ('WRONG_USERNAME','The username you have specified does not exist.'),
                                -6: ('WRONG_PASSWORD','The password is incorrect.'),
                                -7: ('NO_USER_ENTERED','Please enter a username.'),
                                -8: ('PASS_REMINDER_SENT','Password reminder sent. Please check your email.'),
                            }
                            
        if self.email and lemon.cgi.has_key('username'):
            self.errorCodes[-6] = ('WRONG_PASSWORD','The password is incorrect. <a href="%s?mode=reminder&username=%s">Email a reminder</a>.'%(self._buildURL(includeQuery=False), lemon.cgi['username']))
    
    def errorToString(self, error, ignore):
        if error in ignore:
            return ''
        return self.errorCodes[error][0]
        
    def errorToDescription(self, error, ignore):
        if error in ignore:
            return ''
        return self.errorCodes[error][1]
    
    def printPage(self, content, title=None):
        if not title:
            title = self.app
        
        print lemon.getHTTPHeader('text/html') + self.htmlPage%{'title':title, 'content':content}
        
    def getStatus(self):
        #~ if self.app:
            
            #~ for app in self.app:
                
                #~ started = self.session._get('AuthStarted', app=app)
                #~ raise Exception((started, app,self.session._get('AuthStarted', app='news'),self.session._get('AuthStarted')))
                #~ if started and self.session.params['name'] <> app:
                    #~ self.session.params['name'] = app
                    
                    #~ break
            
        started = self.session._get('AuthStarted')
        if started:
            # Checks the expired and idled times
            expireLength = self.session._get('AuthExpireLength')
            accessed     = self.session._get('AuthAccessed')
            idleLength   = self.session._get('AuthIdleLength')
            currentTime  = str( int(time.time()) )
            
            if (expireLength + started) > currentTime:
                return -1
                
            if (idleLength + accessed) > currentTime:
                return -2
            else:
                # Auth session is still valid
                # Retrieve the information for the User
                userName = self.session._get('AuthUser')
                
                # Check that user exists
                
                rows = self.driver.getInfo(self.app, userName)
                if not rows:
                    return -3 # XXX Session is invalid
                self.user['username'] = rows[0][0]
                self.user['password'] = rows[0][1]
                self.user['level'] = rows[0][2]
                self.user['firstname'] = rows[0][3]
                self.user['surname'] = rows[0][4]
                self.user['email'] = rows[0][5]
                return self.user['level']
        
        return -4
                
    def isValid(self):
        "Checks to see if their is a valid entry for the session."
        if self.email and lemon.cgi.has_key('mode') and lemon.cgi.has_key('username') and lemon.cgi['mode'] == 'reminder':
            if lemon.cgi.has_key('email'):
                email = lemon.cgi['email']
                if email == self.driver.getEmail(lemon.cgi['username']):
                    
                    firstname = self.driver.getFirstname(lemon.cgi['username'])
                    if not firstname:
                        firstname = 'user'
                    
                    dict = {
                            'firstname' :   firstname,
                            'system'    :   self.email['system'], 
                            'password'  :   self.driver.getPassword(lemon.cgi['username']),
                            'sender'    :   self.email['sender'],
                        }
                        
                    
                    msg = self.emailMessage%dict
                    
                        
                    if self.email['method'] == 'sendmail':
                        lemon.mail(msg=msg, to=email, subject='Password Reminder', method='sendmail', sendmail=self.email['sendmail'], reply=self.email['reply'], sender=self.email['sender'])
                    elif self.email['method'] == 'smtp':
                        lemon.mail(msg=msg, to=email, subject='Password Reminder', method='smtp', smtp=self.email['smtp'],         reply=self.email['reply'], sender=self.email['sender'])
                    
                    self.printPage(
                        self.getLoginForm(
                            username=lemon.cgi['username'],
                            problem=-8
                        ),
                        'Please Login'
                    )
                else:
                    self.printPage(
                        self.getReminderForm(
                            stickyData={
                                'username':lemon.cgi['username'],
                                'mode':'reminder'
                            },
                            problem='<p>Email address not found. Please enter the email address your account was registered with.</p>'
                        )
                    )
            else:
                self.printPage(
                    self.getReminderForm(
                        stickyData={
                            'username':lemon.cgi['username'],
                            'mode':'reminder'
                        }
                    )
                )
        else:
            status = self.getStatus()
            
            if status > 0:
                return status
                
            elif status == 0:
               
                if self.accessDenied:
                    self.printPage(self.accessDenied, 'Access Denied')
                else:
                    self.printPage(self.getLoginForm(status, self.user['username']),'Please Login')
            else:
                if self.autoLogin:
                    if lemon.cgi.has_key('username'):
                        if lemon.cgi.has_key('password'):
                            status = self.login(lemon.cgi['username'],  lemon.cgi['password'])
                            if status > 0:
                                return status
                            
                            else:
                                if self.autoLogin:
                                    self.printPage(self.getLoginForm(status, lemon.cgi['username']), 'Please Login')
                                    return status
                        else:
                            self.printPage(self.getLoginForm(-6, lemon.cgi['username']), 'Please Login')
                    else:
                        self.session._set('AuthRedirectURL', self._buildURL() )
                        self.printPage(self.getLoginForm(-7,''), 'Please Login')
                else:
                    return -7


            
    def _buildURL(self, includeQuery=True):
        query = ''
        
        if includeQuery and os.environ.has_key('QUERY_STRING') and os.environ['QUERY_STRING']:
            query       = '?' + os.environ['QUERY_STRING']
        
        port        = os.environ['SERVER_PORT']
        host        = os.environ['SERVER_NAME']
        scriptName  = os.environ['SCRIPT_NAME']
        
        if port == '80':
            port = ''
        else:
            port = ':'+str(port)
        
        # "http://%s%s"%(host, port,)
        return "%s%s" %(scriptName, query)
            
            
    def login(self, username, password):
        "Retrieves the information for the User"
        rows = self.driver.getLoginInfo(username)
        if len(rows) == 0:
            return -5 # Username not found
        else:
            if password == rows[0][1]:
                
                # Sets the Started and Accessed times
                self.session._set('AuthStarted',  int(time.time()))
                self.session._set('AuthAccessed',  int(time.time()))
                self.session._set('AuthIdleLength', self.idle)
                self.session._set('AuthExpireLength', self.expire)
                
                # Sets the User key
                self.session._set('AuthUser',  rows[0][0])

                if self.autoRedirect:
                    if self.redirect:
                        redirect = self.redirect
                    elif self.includeQuery:
                        redirect = self.session._get('AuthRedirectURL')
                        self.session._delete('AuthRedirectURL')
                    else:
                        redirect = self._buildURL(includeQuery=False)
                    location = "Location: "+ redirect + '\n\n' # XXX Even if redirect == ''
                    
                    # Problems with the lemon webserver.
                    print location #+ "Content-Length: 343\nKeep-Alive: timeout=15, max=97\nConnection: Keep-Alive\nContent-Type: text/html; charset=iso-8859-1\n\n"
                    #~ print """
                        #~ <html> 
                        #~ <head> 
                        #~ <meta HTTP-EQUIV="refresh" content="10;URL=%s"> 
                        #~ <TITLE>Login Redirect</TITLE> 
                        #~ </HEAD> 
                        #~ <BODY>Login successful. You should be redirected <a href="%s">here</a>.</BODY> 
                        #~ </HTML> 
                        #~ """%(redirect, redirect)
 
                    #~ return self.getStatus() # XXX Should be redirected.self.getStatus()
                
                return self.getStatus()
            else:
                return -6 # Wrong password

    def logout(self):
        """Checks to see if their is a valid entry for the session
        # Deletes the session FOR THIS APP ONLY."""

        if self.session._get('AuthUser'):
            self.session._delete('AuthExpireLength')
            self.session._delete('AuthIdleLength')
            self.session._delete('AuthStarted')
            self.session._delete('AuthAccessed')
            self.session._delete('AuthRedirectURL')
            self.session._delete('AuthUser')
            return 1
        else:
            return 0

    def getUser(self):
        return self.user
        
    def getLoginForm(self, problem=None, username='', stickyData={}):
        "Displays a form allowing the user to login."
        hidden = ''
        for k,v in stickyData.items():
            hidden += '<input type="hidden" name="%s" value="%s"/>'%(k,v)
        dict =  {
                    'name':'login',
                    'method':'post',
                    'action':os.environ['SCRIPT_NAME'],
                    'username':username,
                    'stickyData': lemon.dictAsHiddenFields(self.stickyData), 
                    'error': self.errorToDescription(problem,ignore=[-4,None]),
                }
        
        return self.loginForm % dict
        
        
    def getReminderForm(self, stickyData={}, problem=''):
        "Displays a form allowing the user to login."
        hidden = ''
        for k,v in stickyData.items():
            hidden += '<input type="hidden" name="%s" value="%s"/>'%(k,v)
        dict =  {
                    'name':'login',
                    'method':'post',
                    'action':os.environ['SCRIPT_NAME'],
                    'stickyData': hidden, 
                    'problem':problem,
                }
        
        return self.reminderForm % dict