"""Wraps DB-2.0 API cursors and extends functionality.

Version:

    File Version 0.1 for LEMON 0.2
    
Description:
    
    Module containing the base cursor from which the following cursors are defined:
        lemon.db.mysql
        lemon.db.gadfly
        lemon.db.odbc
"""

__author__ = "James Gardner [email: james at xecos.com]"
__date__ = "02 July 2003"
__version__ = "Version 0.1 for LEMON 0.2"[-3:]
__credits__ = "Guido van Rossum, for an excellent programming language."

import lemon.db
#import lemon.external.dtuple as dtuple
import dtuple, sys

class baseConnection:
    def __init__(self):
        'Set up the database connection'
        raise lemon.db.DBError('This function should be overridden in the derived class to set the self.connection object to a valid DB 2.0 API Connection object.')
        
    def cursor(self, autoExecute=True, fetchMode='tuple', autoConvert=False, convertMode='table', colTypesName='ColTypes', types={}):
        "Return our custom cursor rather than the default connection cursor."
        raise lemon.db.DBError('This function should be overridden in the derived class to return a custom cursor object.')

    # Delegate unrecognised methods/attributes to the cursor object
    def __getattr__(self, name):
        "Delegate unrecognised methods/attributes to the cursor object."
        return getattr(self.connection, name)

class baseCursor:
    "Base class for database adaptation layer"

#
# Initialisation
#
    def setupOptions(self):
        return [
                    [True, False],              # autoExecute
                    ['tuple','dict','dtuple'],  # fetchMode
                    [True, False],              # autoConvert
                    ['table', 'description'],   # convertMode
                ]
                
    def __init__(self, cursor, autoExecute, fetchMode, autoConvert, convertMode, colTypesName='ColTypes', types={} ):  
                
                
        # Verify the options
        
        options = self.setupOptions()
       
        self.autoExecuteOptions = options[0]
        self.fetchModeOptions   = options[1]
        self.autoConvertOptions = options[2]
        self.convertModeOptions = options[3]
        
        
        
        if autoExecute not in self.autoExecuteOptions:
            raise lemon.db.DBError("autoExecute must be one of %s."%(str(self.autoExecuteOptions),))
        if fetchMode not in self.fetchModeOptions:
            raise lemon.db.DBError("fetchMode must be one of %s."%(str(self.fetchModeOptions),))
        if autoConvert not in self.autoConvertOptions:
            raise lemon.db.DBError("autoConvert must be one of %s."%(str(self.autoConvertOptions),))
        if convertMode not in self.convertModeOptions:
            raise lemon.db.DBError("convertMode must be one of %s."%(str(self.convertModeOptions),))
        
        # XXX Verify types as well.
            
        self.cursor = cursor
        
        # Setup the options
        self.colTypesName   = colTypesName
        self.autoExecute    = autoExecute
        self.fetchMode      = fetchMode
        self.types          = types
        self.autoConvert    = autoConvert
        self.convertMode    = convertMode
        self.fieldTypes     = None          # Defined in each derived class
        
        
        # Needed for Gadfly
        self.table = None
        self.fields = None
        self.descrip = None

        self.debug = ''
        self.setupFields()
        
        if self.autoConvert not in self.autoConvertOptions:
            raise Exception("'%s' is not a valid option for autoConvert.")
            
        if self.convertMode == 'table':
            if not self.tableExists(self.colTypesName):
                self.createColTypesTable()
        
    def setupFields(self): # Override
        "This function is overridden in derived classes. It is used to specify the column type integers and strings for each field type Lemon allows. You should also set the self.autoConvertOptions list in this function if it needs to be changed."
        raise DBError("The function 'setupFields(self)' should be overriden in the derived class.")

#
# Delegate unrecognised methods/attributes to the cursor object
#

    def __getattr__(self, name):
        "Delegate unrecognised methods/attributes to the cursor object."
        return getattr(self.cursor, name)

#
# Useful SQL related functions
# 

    def execute(self, sql, values=None):
        self.debug += "SQL:\n%s\n\nTypes:\n%s\n\n\n"%(sql, self.types)
        if values:
            self.cursor.execute(sql,values)
        else:
            self.cursor.execute(sql)

    def tableExists(self, table):
        try:
            self.execute("SELECT * FROM "+table+" WHERE 3=4")
            self.fetchall()
        except:
            return False
        else:
            return True
        
    def columnExists(self, table, column):
        try:
            self.execute("SELECT "+column+" FROM "+table+" WHERE 3=4")
            self.fetchall()
        except:
            return False
        else:
            return True

    def getColumnNames(self, table):
        if self.convertMode == 'table':
            # Returns a list of all the column names for the table 'table'. Used if autoConvert='table'.
            self.execute("select ColumnName from "+self.colTypesName+" where TableName='"+table+"'")
            rows = self.fetchall()
            res = []
            for row in rows:
                res.append(row[0])
            
        elif self.convertMode == 'description':
            self.execute("SELECT * FROM "+table+" WHERE 3=4")
            rows = self.fetchall()
            res = []
            for row in self.description:
                res.append(row[0])
        else:
            raise lemon.db.DBError("convertMode should be 'table' or 'description'.. if it isn't you shouldn't have seen this anyway so this message indicates a bug.")
        return res
        
#
# Table column types information
#

    def createColTypesTable(self):
        "Create the ColumnTypes table required by the DB layer."
        self.execute("CREATE TABLE %s (TableName %s, ColumnName %s, ColumnType %s)"%(self.colTypesName, self.fieldTypes[self.STRING], self.fieldTypes[self.STRING], self.fieldTypes[self.INTEGER]))
        self.execute("insert into "+self.colTypesName+" (TableName,ColumnName,ColumnType) values ('"+self.colTypesName+"', 'TableName', " +str(self.STRING[0])+")") 
        self.execute("insert into "+self.colTypesName+" (TableName,ColumnName,ColumnType) values ('"+self.colTypesName+"', 'ColumnName', " +str(self.STRING[0])+")")
        self.execute("insert into "+self.colTypesName+" (TableName,ColumnName,ColumnType) values ('"+self.colTypesName+"', 'ColumnType', " +str(self.INTEGER[0])+")")
        return True
        
    def addColumnType(self, table, col, c):
        "Add a column type definition to the ColumnTypes table."
        self.execute("select ColumnName, ColumnType from "+self.colTypesName+" where TableName='"+table+"' and ColumnName='"+col+"'")
        rows = self.fetchall()
        if not rows:
            return self.execute("insert into "+self.colTypesName+" (TableName,ColumnName,ColumnType) values ('"+str(table)+"', '"+str(col)+"', " +str(c)+")")
        elif rows[0][1] <> type:
            return self.execute("update "+self.colTypesName+" set ColumnType='"+str(c)+"' where TableName='"+table+"' and ColumnName='"+col+"'")
            
    def removeColumnType(self, table, col):
        "Remove a column type definition to the ColumnTypes table."
        return self.DELETE(self.colTypesName, WHERE="TableName='"+table+"' and ColumnName='"+col+"'")

    def removeTableTypes(self, table):
        "Remove all definitions for the table 'table' from the ColumnTypes table."
        self.execute("delete from "+self.colTypesName+" where TableName='"+table+"'")

#
# Encoding (and error checking # XXX ) Functions 
#

    def checkTypes(self, table, cols):
        "Checks to see if the columns specified by 'cols' are all present in the cache."
        if self.types.has_key(table.upper()):
            for k in cols:
                if not self.types[table.upper()].has_key(k):
                    return False
            return True
        else:
            return False
            
    def mergeTypes(self, table, types):
        if type(table) <> type('l'):
            raise Exception("Bad type '%s' for table name. Should be string."%table)
        if self.types.has_key(table.upper()):
            for k,v in types.items():
                if type(k) <> type('s'):
                    raise Exception("Bad type '%s' for entry into the types cache. Should be string."%k)
                if type(v) <> type(1):
                    raise Exception("Bad type '%s' for entry into the types cache. Should be int."%v)
                if self.types[table.upper()].has_key(k):
                    if self.types[table.upper()][k] <> v:
                        raise Exception("The type information specified doesn't match the data in the types cache.")
                else:
                    self.types[table.upper()][k] = v              
        else:
            self.types[table.upper()] = {}
            for k,v in types.items():
                if type(k) <> type('s'):
                    raise Exception("Bad type '%s' for entry into the types cache. Should be string."%k)
                if type(v) <> type(1):
                    raise Exception("Bad type '%s' for entry into the types cache. Should be int."%v)
                self.types[table.upper()][k] = v
            
            
    def getTableData(self, table): # XXX really might as well get the whole lot in one go.
        "Called to retrieve information on the types of each column for a particular database using the retrieval method set with the autoConvert option."
        if self.autoConvert:
            if self.convertMode == 'table':
                # Returns a list of all the column names for the table 'table'. Used if autoConvert='table'.
    
                self.execute("select ColumnName, ColumnType from "+self.colTypesName+" where TableName='"+table+"'")
                rows = self.fetchall()
                #print rows, '\n\n'
                types = {}
                for row in rows:
                    types[row[0].upper()]=int(row[1])
                self.mergeTypes(table, types)
                
            elif self.convertMode == 'description':
                if not self.types.has_key(table):
                    # Obtain the type from the database
                    self.execute("SELECT * FROM "+table+" WHERE 3=4")
                    types = {}
                    for row in self.description:
                        types[row[0].upper()]=int(row[1])
                    self.mergeTypes(table, types)
            else:
                 raise lemon.db.DBError("You cannot retrieve table data without setting convertMode to 'table' or 'description'.")
        else:
            raise lemon.db.DBError("You cannot retrieve table data without setting autoConvert to True.")
        return self.types[table.upper()]
        

    def getTypes(self, table, fields):
        "Called to get a list of field types from a table name and a list of field names."
        if self.checkTypes(table, fields):
            desc = self.types[table.upper()]
        else:
            desc = self.getTableData(table)
        
        colTypes = []
        for col in fields: # ie for each column
            colTypes.append(desc[col.upper()])
        return colTypes
 
        
    def encodeRow(self, types, values):
        "Called to encode the data sent as SQL to ensure all strings are handled correctly across each database."
        if len(values) != len(types):
            raise lemon.db.DBError("Type conversion error: The values submitted don't match the types values from the database. %s %s"%(types, values))
       
        res = []
        for field in range(len(values)):
            Type = types[field]
            
            # Begin conversion
            
            # String Types
            #import pdb; pdb.set_trace()
            if Type in self.CHAR:
                if values[field] == None:
                    res.append(None)
                else:
                    char = str(values[field])[0:1]
                    if char <> values[field]:
                        raise lemon.db.DBError('Fields of type CHAR should only be strings one character long.')
                    res.append(self.encodeChar(values[field]))
                
            elif Type in self.STRING:
                if values[field] == None: # XXX Allows None
                    res.append(None)
                else:
                    stri = str(values[field])[:255]
                    if stri <>  str(values[field]):
                        raise lemon.db.DBError('Fields of type STRING should only be strings between 1 and 255 characters long.')
                    res.append(self.encodeString(values[field]))
               
            # Binary Types
            elif Type in self.BINARY:
                if values[field] == None: # XXX Allows None
                    res.append(None)
                else:
                    res.append(self.encodeBinary(values[field]))

            # Number Types
            elif Type in self.INTEGER:
                if values[field] == None: # XXX Allows None
                    res.append(None)
                else:
                    try:
                        int_ = int(values[field])
                    except:
                        raise lemon.db.DBError("Value '%s' is not an integer. It is type '%s'"%(values[field], str(type(values[field]))))
                    if int_ is not values[field]:
                        raise lemon.db.DBError("Value '%s' is not an integer. It is type '%s'"%(values[field], str(type(values[field]))))
                    res.append(self.encodeInteger(values[field]))

            elif Type in self.FLOAT: # XXX Points to Note: Doesn't necessarily keep decimal places.
                if values[field] == None: # XXX Allows None
                    res.append(None)
                else:
                    int_ = float(values[field])
                    if int_ <>  values[field]:
                        raise lemon.db.DBError('Fields of type FLOAT should be valid python floats.')
                    res.append(self.encodeFloat(values[field]))

            # Date Types
            elif Type in self.DATE:
                if values[field] == None:
                    res.append(None)
                else:
                    res.append(self.encodeDate(values[field]))
            elif Type in self.TIME:
                if values[field] == None:
                    res.append(None)
                else:
                    res.append(self.encodeTime(values[field]))
            elif Type in self.DATETIME:
                if values[field] == None:
                    res.append(None)
                else:
                    res.append(self.encodeDateTime(values[field]))
            else:

                raise lemon.db.DBError("Warning: No conversion method has been implemented for fields of type '%s' for '%s'"%(Type,values[field]))
                #else:
                #    res.append(values[field])   
                
        return tuple(res)

    def encodeString(self, field):
        return "'"+str(field).replace("'","''")+"'"

    def encodeChar(self, field):
        return "'"+str(field)[0:1].replace("'","''")+"'"    

    def encodeBinary(self, field):
        return "'"+str(field).replace("'","''")+"'"
        
    def encodeInteger(self, field):
        try:
            return int(field)
        except:
            pass
        else:
            raise Exception("'%s' is not an int."%field)
        
    def encodeFloat(self, field):
        return float(field)

    def encodeDate(self, field):
        return "'"+str(field)+"'"

        
    def encodeTime(self, field):
        return "'"+str(field)+"'"
        
    def encodeDateTime(self, field):
        return "'"+str(field)+"'"
        
    def decodeRow(self, types, values):
        "Called to convert the results of a select or function call to the correct types."
        if len(types) != len(values):
            raise lemon.db.DBError("Type conversion error: The types of results from the database don't match the type descriptions from the database.")
        
        res = []
        for field in range(len(values)):
            
            # Error check the types
            Type = types[field]
                
            # Begin conversion

            # String Types
            if Type in self.CHAR:
                if values[field] == None:
                    res.append(None)
                else:
                    res.append(self.decodeChar(values[field]))
            elif Type in self.STRING:
                if values[field] == None:
                    res.append(None)
                else:
                    res.append(self.decodeString(values[field]))  

            # Binary Types
            elif Type in self.BINARY:
                if hasattr(values[field], 'tostring'):
                    res.append(self.decodeBinary(values[field].tostring()))
                elif values[field] == None:
                    res.append(None)
                else:
                    res.append(self.decodeBinary(values[field]))
                
            # Number Types
            elif Type in self.INTEGER:
                if values[field] == None:
                    res.append(None)
                else:
                    res.append(self.decodeInteger(values[field]))
            elif Type in self.FLOAT: # XXX Points to Note: Doesn't necessarily keep decimal places.
                if values[field] == None:
                    res.append(None)
                else:
                    res.append(self.decodeFloat(values[field])) 

            # Date Types
            elif types[field] in self.DATE:
                if values[field] == None:
                    res.append(None)
                else:
                    res.append(self.decodeDate(values[field]))
            elif types[field] in self.TIME:
                if values[field] == None:
                    res.append(None)
                else:
                    res.append(self.decodeTime(values[field]))
            elif types[field] in self.DATETIME:
                if values[field] == None:
                    res.append(None)
                else:
                    res.append(self.decodeDateTime(values[field]))
            else:
                #if self.showWarnings == True:
                raise lemon.db.DBError("Warning: No conversion method has been implemented for fields of type '%s'"%types[field])
                #else:
                #    res.append(values[field])
        return tuple(res)

    def decodeString(self, field):
        #if type(field) is type([]) and len(field) == 2:
        #    return str(field[1]) # XXX  XXXXXXXXXXXXXXXXXXXXXXXXxxx James's quick fix!
        #else:
        return str(field)

    def decodeChar(self, field):
        return str(field)[0:1]            
                            
    def decodeBinary(self, field):
        return field
        
    def decodeInteger(self, field):
        try:
            return int(field)
        except ValueError:
            raise Exception("Invalid value '%s' for integer."%field)
        
    def decodeFloat(self, field):
        try:
            return float(field)
        except ValueError:
            raise Exception("Invalid value '%s' for float."%field)
            
    def decodeDate(self, field):
        tup = lemon.date.isodate2tuple(str(field))
        return lemon.date.date(tup[0],tup[1],tup[2])
        
    def decodeTime(self, field):
        if len(str(field))<8:
            field = '0'+str(field)
            
        tup = lemon.date.isotime2tuple(str(field))
        return lemon.date.time(tup[3],tup[4],tup[5])
        
    def decodeDateTime(self, field):
        tup = lemon.date.isodatetime2tuple(str(field))
        return lemon.date.datetime(tup[0],tup[1],tup[2],tup[3],tup[4],tup[5])

#
# Fetch operations
#   
        
    def fetchRows(self, fetchMode=None, types=None):
        """Retrieve all the rows from the query in the format of a sequence of results.
        
        The 'types' value should be a dictionary containing the table names with each table being the key to another dict containg the column names and types."""
        
        if fetchMode <> None:
            if fetchMode not in self.fetchModeOptions:
                raise lemon.db.DBError("fetchMode must be one of %s."%(str(self.fetchModeOptions),))
        else:
            fetchMode = self.fetchMode
        if types:
            for k in types.items():
                self.mergeTypes(k, types[k])
            
        returnRows = []
        results = self.fetchall()
        if len(results) == 0:
            return results

        colTypes = []
        colNames = []
        #raise Exception(self.fields)
        
        for colnum in range(len(results[0])): # ie for each column
            #print self.fields[colnum]
            colNames.append(self.fields[colnum])
            
            #self.description[colnum][1]=self.types[self.table][self.description[colnum][0]]
            colTypes.append(self.types[self.table.upper()][self.description[colnum][0].upper()])
            
        rows = []
        for row in results:
            r = row
            if self.autoConvert == True:
                r = self.decodeRow(colTypes,row)
            if fetchMode == 'dict':
                dict={}
                for field in range(len(r)):
                    dict[colNames[field]] = r[field]
                rows.append(dict)
            elif fetchMode == 'dtuple':
                descr = dtuple.TupleDescriptor([[n] for n in colNames])
                rows.append(dtuple.DatabaseTuple(descr, row))
            elif fetchMode == 'tuple':
                rows.append(r)
            else:
                raise Exception("'%s' is not a valid option for fetchMode.")
        return tuple(rows)

#
# Data Manipulation language SQL statement generators
#

    def select(self, fields, table=None, where=None, orderby=None, rest=None, types=None, autoConvert=None, autoExecute=None, fetchMode=None):
        """
        fields should be either:
            1. A string
            2. A list of strings
            3. A list of strings and tuples of length 2 in the form (sql, alias)
        
            If a single list contaning two enties is passed it is taken to be 2. not 3.
        
        table should be either:
            1. The table name as a string
            2. A list of table names
            
        WHERE should be a valid where clause
        
        ORDERBY should be a valid ORDER BY string
        """

        # Setup the default actions

        if autoConvert <> None:
            if autoConvert not in self.autoConvertOptions:
                raise lemon.db.DBError("autoConvert must be one of %s."%(str(self.autoConvertOptions),))
        else:
            if types:
                raise Exception("You cannot specify types as no conversion is taking place becuase autoConvert is False.")
            autoConvert = self.autoConvert
        if autoExecute <> None:
            if autoExecute not in self.autoExecuteOptions:
                raise lemon.db.DBError("autoExecute must be one of %s."%(str(self.autoExecuteOptions),))
        else:
            autoExecute = self.autoExecute
        if not autoExecute and fetchMode:
            raise Exception("You cannot set the fetchMode if you haven't set autoExecute as there would be nothing to fetch.")
        if fetchMode <> None:
            if fetchMode not in self.fetchModeOptions:
                raise lemon.db.DBError("fetchMode must be one of %s."%(str(self.fetchModeOptions),))
        else:
            fetchMode = self.fetchMode


        typesDict = None
        if types:
            typesDict = {table:types}
                
                
        # Get the fields in the correct format
        if fields == '*':
            fields = self.getColumnNames(table)
            
        if type(fields) is type(''):
           fields = [fields]
        
        self.table = table
        self.fields = fields
        
        sqlFields = []
        for f in fields:
            if type(f) is type(''):
                sqlFields.append(f)
            else:
                raise Exception(f)
                sqlFields.append(self.fieldAlias(f))

        sql = "SELECT " + ", ".join(sqlFields)
        
        if table:
            sql += " FROM %s" % self.tables(table)
        if where:
            sql += " WHERE %s" % where
        if orderby:
            sql += " ORDER BY %s" % orderby
        if rest:
            sql += " %s" % rest
            
        # Don't need to worry about convertQuery since there are no values    
            
        if autoExecute == False:
            return sql
        else:
            # Don't need to worry about convertResult since it is taken care of in fetchRows()
            # Don't need to worry about fetchMode since it is taken care of in fetchRows()
            
            if autoConvert == True:
                if not self.checkTypes(table, fields):
                    self.getTableData(table)
            self.execute(sql)
            return self.fetchRows(fetchMode, typesDict)

    def update(self, table, fields, values, where=None, types=None, autoConvert=None, autoExecute=None):
        
        # Setup the default actions
        if autoConvert <> None:
            if autoConvert not in self.autoConvertOptions:
                raise lemon.db.DBError("autoConvert must be one of %s."%(str(self.autoConvertOptions),))
        else:
            autoConvert = self.autoConvert
        if autoExecute <> None:
            if autoExecute not in self.autoExecuteOptions:
                raise lemon.db.DBError("autoExecute must be one of %s."%(str(self.autoExecuteOptions),))
        else:
            autoExecute = self.autoExecute

        if type(fields) <> type(()) and type(fields) <> type([]):
            fields=(fields,)
        if type(values) <> type(()) and type(values) <> type([]):
            values = tuple([values])
            
        if len(fields) != len(values):
            raise lemon.db.DBError("UPDATE column count disqlFieldsers from value count.")
    
        # Convert the values to the correct type
        if autoConvert == False:
            for v in values:
                if type(v) <> type(''):
                    raise Exception("If you don't want values autoConverted they must be supplied as properly quoted strings.")
        else:
            if types:
                self.mergeTypes(table, types)
            types = self.getTypes(table, fields)
            values = self.encodeRow(types, values)
        
        params=[]
        for item in values:
            params.append('%s')

        sql = "UPDATE %s SET %s" % (table, ", ".join(map(lambda x,y: str(x)+"="+str(y), fields, params)))

        if where:
            sql += " WHERE " + where
        val = []
        for v in values:
            if v == None:
                val.append('NULL')
            else:
                val.append(v)
        sql = sql%(tuple(val))
        if autoExecute == True:
            self.execute(sql)
        return sql

    def insert(self, table, fields, values, types=None, autoConvert=None, autoExecute=None):
        
        # Setup the default actions
        if autoConvert <> None:
            if autoConvert not in self.autoConvertOptions:
                raise lemon.db.DBError("autoConvert must be one of %s."%(str(self.autoConvertOptions),))
        else:
            autoConvert = self.autoConvert
        if autoExecute <> None:
            if autoExecute not in self.autoExecuteOptions:
                raise lemon.db.DBError("autoExecute must be one of %s."%(str(self.autoExecuteOptions),))
        else:
            autoExecute = self.autoExecute

        fieldList = ""
        valuesString = ""
        if type(fields) <> type(()) and type(fields) <> type([]):
            fields = tuple([fields])
        fieldList = ", ".join(fields)

        if len(fields) is 1:
            valuesString = '%s'
        else:
            valuesString = (('%s' + ", ")*(len(fields)-1))+'%s'

        sql = "INSERT INTO %s (%s) VALUES (%s)" % (table, fieldList, valuesString)

        # Convert the values to the in the correct format
        if type(values) <> type(()) and type(values) <> type([]):
            values = tuple([values])
            
        # Convert the values to the correct type
        if autoConvert == False:
            for v in values:
                if type(v) <> type(''):
                    raise Exception("If you don't want values autoConverted they must be supplied as properly quoted strings.")
        else:
            if types:
                self.mergeTypes(table, types)
            types = self.getTypes(table, fields)
            values = self.encodeRow(types, values)
            
        # Execute the query
        val = []
        for v in values:
            if v == None:
                val.append('NULL')
            else:
                val.append(v)
        sql = sql%(tuple(val))
        if autoExecute is True:
            self.execute(sql)
        return sql

    def delete(self, table, where=None, autoExecute=None):
        # Setup the default actions
              
        if autoExecute <> None:
            if autoExecute not in self.autoExecuteOptions:
                raise lemon.db.DBError("autoExecute must be one of %s."%(str(self.autoExecuteOptions),))
        else:
            autoExecute = self.autoExecute
  
        sql = "DELETE FROM %s" % table
        if where:
            sql += " WHERE %s" % where
        if autoExecute == True:
            self.execute(sql)
        return sql

    def create(self, table, fields, autoExecute=None):#, auto=None, primary=None, notNULL=None, unique=None, default=None):
        """Fields:
            (name, type, notNULL, default, auto, primary, unique)
        """

        if autoExecute <> None:
            if autoExecute not in self.autoExecuteOptions:
                raise lemon.db.DBError("autoExecute must be one of %s."%(str(self.autoExecuteOptions),))
        else:
            autoExecute = self.autoExecute
            
        if table == self.colTypesName:
            raise lemon.db.DBError("You cannot create the '"+self.colTypesName+"' table using this method. It is a reserved table name.")
        
        self.types[table.upper()] = {}
        
        for f in fields:
            if type(f[1]) in [type([]), type((1,))]:
                c = f[1][0]
            else:
                c = f[1]
            self.types[table.upper()][f[0].upper()] = c
            if self.autoConvert == True and self.convertMode == 'table':
                self.addColumnType(table, f[0], c)

        sqlFields = []
        for f in range(len(fields)): 
            if len(fields[f][0])>255:
                raise lemon.db.DBError('Column names must be 255 charaters or less.')
                
            # name, type
            stri = "%s %s" % (fields[f][0], self.fieldTypes[fields[f][1]])
            # notNULL
            if len(fields[f])>2:
                if fields[f][2] == True:
                    stri += ' NOT NULL'
            # default
            if len(fields[f])>3:
                if fields[f][3]:
                    stri += ' DEFAULT '+str(fields[f][3])
            # auto
            if len(fields[f])>4:
                if fields[f][4] == True:
                    stri += ' AUTO_INCREMENT'
            # primary
            if len(fields[f])>5:
                if fields[f][5] == True:
                    stri += ' PRIMARY KEY'
            # unique
            if len(fields[f])>6:
                if fields[f][6] == True:
                    stri += ' UNIQUE'
            
            sqlFields.append(stri)
        sql = "CREATE TABLE %s (%s)" % (table, ", ".join(sqlFields))
        
        if autoExecute == True:
            self.execute(sql)
        return sql
            
    def alter(self, table, mode, columnName, columnType=None, autoExecute=None):
        
        if autoExecute <> None:
            if autoExecute not in self.autoExecuteOptions:
                raise lemon.db.DBError("autoExecute must be one of %s."%(str(self.autoExecuteOptions),))
        else:
            autoExecute = self.autoExecute
            
        if table == self.colTypesName:
            raise lemon.db.DBError("You cannot alter the '"+self.colTypesName+"' table using this method. It is a reserved table name.")
    
        if mode not in ['add','drop']:
            raise lemon.db.DBError("The 'mode' attribute can only be 'add' or 'drop'.")
    
        if mode == 'add':
            if not columnType:
                raise lemon.db.DBError("You must specify a columnType.")
                
            
            if not self.types.has_key(table.upper()):
                self.types[table.upper()] = {}
            self.types[table.upper()][columnName.upper()] = columnType
            if self.autoConvert == True and self.convertMode == 'table':
                if type(columnType) in [type([]), type((1,))]:
                    d = columnType[0]
                else:
                    d = columnType
                self.addColumnType(table, columnName, d)
            
            sql = "ALTER TABLE %s ADD %s %s" % (table, columnName, self.fieldTypes[columnType])
            
        else:
            del self.types[table.upper()][columnName.upper()]
            if self.autoConvert == True and self.convertMode == 'table':
                self.removeColumnType(table, columnName)
            sql = "ALTER TABLE %s DROP %s" % (table, columnName)
            
    
        if autoExecute == True:
            self.execute(sql)
        return sql
            
    def drop(self, table, autoExecute=None):
        "Remove a table from the database."

        if autoExecute <> None:
            if autoExecute not in self.autoExecuteOptions:
                raise lemon.db.DBError("autoExecute must be one of %s."%(str(self.autoExecuteOptions),))
        else:
            autoExecute = self.autoExecute
            
        if table == self.colTypesName:
            raise lemon.db.DBError("You cannot drop the '"+self.colTypesName+"' table using this method. It is a reserved table name.")
        if self.types.has_key(table.upper()):
            del self.types[table.upper()]
        if self.autoConvert == True and self.convertMode == 'table':
            self.removeTableTypes(table)

        result = "DROP TABLE %s" % table
        if autoExecute is True:
            self.execute(result)
        return result
            
    def max(self, field, table, where=None):
        return self.function('max',field, table, where)
        
    def min(self, field, table, where=None):
        return self.function('min',field, table, where)
        
    def sum(self, field, table, where=None):
        return self.function('sum',field, table, where)
        
    def average(self, field, table, where=None):
        return self.function('avg',field, table, where)

#
# Utility functions: may be overridden to modify SQL grammar
#

    def function(self, func, field, table, where=None):
        """Warning: Please note.. no conversion is done here."""
        if func.upper() in ['MAX', 'MIN', 'SUM', 'AVG']:
            self.table = table
            self.fields = field
            #result = "SELECT %s(%s) AS %s FROM %s" % (func, field, field, table)
            result = "SELECT %s(%s) FROM %s" % (func, field, table)
            if where:
                result += " WHERE %s" % where
                
            if self.autoExecute is True:
                self.execute(result)
                val = self.fetchall()[0][0]
                return val
            else:
                return result
        else:
            raise lemon.db.DBError("The function '%s' is not supported by the lemon.db module."%func)
            
    def fieldAlias(self, f):
        # XXX Joins don't work in Gadfly
        if len(f) != 2:
            raise ValueError, "Field alias should be exactly two elements"
        else:
            return " AS ".join(f)

    def tables(self, FROM):
        # XXX Joins don't work in Gadfly
        if type(FROM[0]) is type(''):
            FROM = [FROM]
        result = []
        for t in FROM:
            if type(t) is type(''):
                result.append(t)
            else:
                result.append(self.tblalias(t))
        return ", ".join(result)

    def tblalias(self, t):
        # XXX Joins don't work in Gadfly
        if len(t) != 2:
            raise ValueError, "Table alias should be exactly two elements"
        else:
            return " AS ".join(t)

#
# Deprecated Methods
#

    def SELECT(self, fields, FROM=None, WHERE=None, ORDERBY=None, REST=None, types=None, autoConvert=None, autoExecute=None, fetchMode=None):
        return self.select(fields, FROM, WHERE, ORDERBY, REST, types, autoConvert, autoExecute, fetchMode)
        
    def UPDATE(self, table, fields, VALUES, WHERE=None, types=None, autoConvert=None, autoExecute=None):
        return self.update(table, fields, VALUES, WHERE, types, autoConvert, autoExecute)
        
    def INSERT(self, table, fields, VALUES, types=None, autoConvert=None, autoExecute=None):
        return self.insert(table, fields, VALUES, types, autoConvert, autoExecute)
        
    def DELETE(self, table, WHERE=None, autoExecute=None):
        return self.delete(table, WHERE, autoExecute)

    def CREATE(self, table, fields):
        return self.create(table, fields)
            
    def ALTER(self, table, mode, columnName, dataType=None):
        return self.alter(table, mode, columnName, dataType)
            
    def DROP(self, table):
        return self.drop(table)
 
    def FUNCTION(self, func, field, table, WHERE=None):
        """Warning: Please note.. no conversin is done here."""
        return self.function(func, field, table, WHERE)
        
      

#~ #
#~ # Change parameter operations (Shouldn't really be needed as you can set them through the functional interface)
#~ #
    #~ # Set
    #~ def setParamStyle(self, style):
        #~ if style in ['?', '%s',]:
            #~ self.paramStyle = style
        #~ else:
            #~ raise lemon.db.DBError("The parameter '%s' is not a valid database parameter for this database."%style)

    #~ def setFetchMode(self, mode):
        #~ if mode in ['dict', 'tuple']:
            #~ self.fetchMode = mode
        #~ else:
            #~ raise lemon.db.DBError("The mode '%s' is not a valid database fetch mode. Use 'dict' or 'tuple'."%mode)
            
    #~ def setAutoExecute(self, mode):
        #~ if mode in [True, False]:
            #~ self.autoExecute = mode
        #~ else:
            #~ raise lemon.db.DBError("Auto execute mode can only be set to True or False.")
            
    #~ def setConvertQuery(self, mode):
        #~ if mode in [True, False]:
            #~ self.convertQuery = mode
        #~ else:
            #~ raise lemon.db.DBError("Auto convert mode can only be set to True or False."%mode)

    #~ def setConvertResult(self, mode):
        #~ if mode in [True, False]:
            #~ self.convertResult = mode
        #~ else:
            #~ raise lemon.db.DBError("Auto convert mode can only be set to True or False."%mode)
            
    #~ def setShowWarnings(self, mode):
        #~ if mode in [True, False]:
            #~ self.showWarnings = mode
        #~ else:
            #~ raise lemon.db.DBError("Supress warnings mode can only be set to True or False."%mode)
            
            
    #~ # Get
    #~ #def getParamStyle(self):
    #~ #    return self.paramStyle
        
    #~ def getFetchMode(self):
        #~ return self.fetchMode
        
    #~ def getAutoExecute(self):
        #~ return self.autoExecute
        
    #~ def getConvertQuery(self):
        #~ return self.convertQuery
        
    #~ def getConvertResult(self):
        #~ return self.convertResult
    #~ #def getShowWarnings(self):
    #~ #    return self.showWarnings
    
    
    
#~ #
#~ # Conversion Functions which should be defined in the module
#~ #

    #~ def Date(self,year,month,day):
        #~ "This function constructs an object holding a date value. "
        #~ return dbmodule.Date(year,month,day)
        
    #~ def Time(self,hour,minute,second):
        #~ "This function constructs an object holding a time value. "
        #~ return dbmodule.Time(hour,minute,second)

    #~ def Timestamp(self,year,month,day,hour,minute,second):
        #~ "This function constructs an object holding a time stamp value. "
        #~ return dbmodule.Timestamp(year,month,day,hour,minute,second)

    #~ def DateFromTicks(self,ticks):
        #~ "This function constructs an object holding a date value from the given ticks value (number of seconds since the epoch; see the documentation of the standard Python time module for details). "
        #~ return dbmodule.DateFromTicks(ticks)

    #~ def TimeFromTicks(self,ticks):
        #~ "This function constructs an object holding a time value from the given ticks value (number of seconds since the epoch; see the documentation of the standard Python time module for details). "
        #~ return dbmodule.TimeFromTicks(ticks)

    #~ def TimestampFromTicks(self,ticks):
        #~ "This function constructs an object holding a time stamp value from the given ticks value (number of seconds since the epoch; see the documentation of the standard Python time module for details). "
        #~ return dbmodule.TimestampFromTicks(ticks)

    #~ def Binary(self,stri):
        #~ "This function constructs an object capable of holding a binary (long) string value."
        #~ return dbmodule.Binary(stri)

    #~ def fetchRow(self):
        #~ "XXX Not Checked Yet... Retrieve a single row from the query."
        #~ returnRows = []
        #~ result = self.fetchone()
        
        
        #~ if self.__class__.__name__ == 'gadflyCursor':
            #~ colNames = self.fields
            #~ colTypes = self.getTypes(self.table, self.fields)
        #~ else:

            #~ colTypes = []
            #~ colNames = []
            
            #~ for colnum in range(len(result)): # ie for each column
                #~ colNames.append(self.description[colnum][0])
                #~ colTypes.append(self.description[colnum][1])

        #~ rows = []
        #~ for row in [result]:
            #~ r = row
            #~ if self.convertQuery == True:
                #~ r = self.decodeRow( colTypes, row)

            #~ if self.fetchMode == 'dict':
                #~ dict={}
                #~ for field in range(len(r)):
                    #~ dict[colNames[field]] = r[field]
                #~ return dict
            #~ else:
                #~ return tuple(r)


#
# Useful stusqlFields!
#

###
### Data definition language SQL statement generators
###
##    def CREATE_INDEX(self, ixname, tbl, fields, UNIQUE=0):
##        if type(fields) is type(''):
##            fields = [fields]
##        if UNIQUE:
##            unique = "UNIQUE "
##        else:
##            unique = ""
##        return "CREATE %sINDEX %s ON %s(%s)" % (unique, ixname, tbl, ", ".join(fields))
##
##    def CREATE_VIEW(self, vname, fields, select):
##        return "CREATE VIEW %s (%s) AS\n    %s" % (vname, ", ".join(fields), select)
##
##    def DROP_INDEX(self, i, t):
##        return "DROP INDEX %s" % i
##
##    def DROP_TABLE(self, t):
##        return "DROP TABLE %s" % t
##
##    def DROP_VIEW(self, v):
##        return "DROP VIEW %s" % v
##
##    def UNION(self, *stmts, **kywds):
##        union = "UNION"
##        if kywds.has_key("DISTINCT") and kywds["DISTINCT"]:
##            union += " DISTINCT"
##        union = "\n    %s\n" % union
##        result = union.join(stmts)
##        if kywds.has_key("REST"):
##            result = "%s\n    %s" % (result, kywds["REST"])
##        return result

  

