"""Lemon DB interface to the gadfly cursor.


THIS SOFTWARE IS PROVIDED WITH ABSOLUTELY NO WARRANTY WHATSOVER. YOU USE IT ENTIRELY AT YOUR OWN RISK.
IT ISN'T EVEN SUPPLIED WITH AN IMPLIED FITNESS FOR ANY PURPOSE.

WARNING:

    This Gadfly cursor is experimental. It works for most purposes but becuase Gadlfy is not a full SQL database
    and is not particularly robust you may come across issues in its use. Gadfly also needs to be connected
    in a different mode for performing DROP, ALTER and CREATE commands (see below). If you do not use this mode
    you could run into problems.
    
    You should learn Gadfly's quirks before using this software and then you are much less likely to run into
    problems.
    
    DO NOT USE IN PRODUCTION SYSTEMS.
    
Version:

    File Version 0.1 for LEMON 0.2
    
Description:
    
    Module containing the gadfly cursor.

Warning: Gadfly has no real typing. Al results are just varchars.

    To create a database using gadfly use the usual interface not the lemon.db one.
    
        import gadfly
        import lemon.db.gadfly
        
        connection = gadfly.gadfly()
        connection.startup("mydatabase", "mydirectory")
        
        cursor = lemon.db.gadfly.gadflyCursor(conn=connection)
        
        cursor.execute("create table ph (nm varchar, ph varchar)")
        cursor.execute("insert into ph(nm, ph) values ('arw', '3367')")
        cursor.setShowWarnings(False)
        for x in cursor.SELECT(('nm', 'ph'),"ph"):
            print x
        print cursor.description
        
        cursor.commit()
        
Making a backup

    WARNING:
    
    Gadfly stores the data in the database as text files. If you are transferring files from a server make sure 
    you set the transfer to 'Binary' as you may well corrupt the files transferring them in 'Auto' or 'Text' mode.
    You have been warned!
"""

__author__ = "James Gardner [email: james at xecos.com]"
__date__ = "02 July 2003"
__version__ = "Version 0.1 for LEMON 0.2"[-3:]
__credits__ = "Guido van Rossum, for an excellent programming language."

import lemon.db
import base


class gadflyConnection(base.baseConnection):
    "Gadfly conncetion object."
    
    def __init__(self, database, dir):
        'Set up the database connection'
        import gadfly, os
        self.connection = gadfly.gadfly(database, os.path.abspath(dir.replace('/',os.sep)))

    def cursor(self, autoExecute=True, fetchMode='tuple', autoConvert=False, convertMode='table', colTypesName='ColTypes', types={}):
        "Return our custom cursor rather than the default connection cursor."
        return gadflyCursor(self.connection.cursor(), autoExecute, fetchMode, autoConvert, convertMode, colTypesName, types)

class gadflyStartupConnection(gadflyConnection):
    "Special Gadlfy mode for setting up a database."
    def __init__(self, database, dir):
        'Set up the database connection'
        import gadfly, os, os.path
        self.connection = gadfly.gadfly()
        self.connection.startup(database, os.path.abspath(dir))
        cursor = self.cursor(autoConvert=True)
        if not cursor.tableExists('ColTypes'): # XXX
            cursor.createColTypesTable()
        self.connection.commit()

class gadflyCursor(base.baseCursor):
    
    def setupOptions(self):
        return [
                    [True, False],              # autoExecute
                    ['tuple','dict','dtuple'],  # fetchMode
                    [True, False],              # autoConvert
                    ['table'],                  # convertMode
                ]

    def setupFields(self):
        """For this to work the definitions must be tuples containing just one integer, otherwise the integer field
        causes problems when adding a column definition"""
        
        self.CHAR       = (1,)
        self.STRING     = (2,)
        self.BINARY     = (3,)
        self.INTEGER    = (4,)
        self.FLOAT      = (5,)
        self.DATE       = (6,)
        self.TIME       = (7,)
        self.DATETIME   = (8,)
    
        self.fieldTypes =  {    
                                self.CHAR       : "varchar",
                                self.STRING     : "varchar",
                                self.BINARY     : "varchar",
                                self.INTEGER    : "integer",
                                self.FLOAT      : "float",
                                self.DATE       : "varchar",
                                self.TIME       : "varchar",
                                self.DATETIME   : "varchar",
                            }
    
    def alter(self, conn, table, mode, columnName, columnType=None, default=None):
        """
        Note: The code has been re-written and seems to work well so the info below may not be relevant.
              Also no commits are used.

        Manually performs a table alteration in Gadfly.
        
        WARNING: The function is not robust and may fail, make sure you read the docstring for the Gadfly ALTER()
                 command and that you make a proper backup as described in the docstring for this module.
                 
        NOTE:    This function has a different syntax to ALTER(). This is because Gadfly doesn't suppot NULLs so you
                 must specify a default value of the correct type for the column.
                 
        ALTER() does not have the same effect in the Gadfly cursor as in other cursors.
        
        Differences:
            1. Gadfly doesn't support ALTER.
                    This means that the implementation has been done manually by reading all the results
                    from the table, deleting the table, creating a new table and putting all the results
                    back in. This is not reliable. several self.commit() calls are made which menas 
                    transactions could behave strangely. Also gadfly tends to like being disconnected from and
                    reconnected to before any further calls are made.
                    
        If you really need to alter a gadfly table, make a backup and then connect to Gadfly in the 
        way you would if you were creating a table for the first time (as described in the gadfly documentation)
        then perform the alter, commit, and disconnect. The table should have been altered successfully.
        """
        # Find a free temporary table name
        counter = 0
        found = True
        while found == True:
            if counter == 10:
                raise Exception("Problem with Gadfly Cursor... 10 temporary tables used in the table ALTER command already exist. This probably means the ALTER function isn't correctly deleting tables. You can comment out this line as a temporary fix and the database will just get full of un-needed tables or try to fix the faulty code.")
            found = self.tableExists('temp_'+table+str(counter))
            counter += 1
        tempTableName = 'temp_'+table+str(counter)
        
        if mode == 'add':
            if not columnType:
                raise lemon.db.DBError("You must specify a columnType.")
            
            if default == None:
                raise lemon.db.DBError("With Gadfly connections You must specify a default value to the ALTER command when adding columns since the database doesn't support NULLs.")
                
            # Get the field names and values
            fields = []
            fieldNames = []
            
            # Do all this manually.
            self.execute("select ColumnName, ColumnType from "+self.colTypesName+" where TableName='"+table+"'")
            rows = self.fetchall()
            for r in rows:
                fields.append((r[0], (r[1],)))                      # name, tuple of the type.
                fieldNames.append(r[0])
                
            # Get current fields
            sql = 'SELECT '+ ', '.join(fieldNames) +' FROM '+ table
            self.execute(sql)
            current = self.fetchall()
        
            # Add the new column to the fields definition
            fields.append((columnName, columnType))   # Adding notNULL = None, default
            
            # Add the new column to the fieldNames list
            fieldNames.append(columnName)

            # Add the default value to each of the rows in the current table.
            newValues = []
            for row in current:
                newRow = list(row)
                newRow.append(default)
                newValues.append(tuple(newRow))
            newValues = tuple(newValues)
            
            self.DROP(table)
            conn.commit()
            self.CREATE(table, fields)
            conn.commit()
            
            string_ = ''
            for i in newValues[0]:
                string_ += ', ?'
            sql = 'INSERT INTO '+table+' ('+', '.join(fieldNames)+') VALUES ('+string_[2:]+')'

            for values in newValues:
                self.execute(sql,values)
            conn.commit()
            
        elif mode == 'drop':

            # Get the field names and values
            # Do all this manually.
            fieldNames = []
            fieldTypes = []
            self.execute("select ColumnName, ColumnType from "+self.colTypesName+" where TableName='"+table+"'")
            rows = self.fetchall()
            for r in rows:
                fieldNames.append(r[0])
                fieldTypes.append(r[1])

            # Get current fields
            sql = 'SELECT '+ ', '.join(fieldNames) +' FROM '+ table
            self.execute(sql)
            current = self.fetchall()
            
            
            # Do all this manually.
            fields = []
            newFieldNames = []
            
            counter = index = 0
            for r in range(len(fieldNames)):
                if fieldNames[r] == columnName:
                    index = counter
                else:
                    fields.append((fieldNames[r], (fieldTypes[r],)))
                    newFieldNames.append(fieldNames[r])                    # name, tuple of the type.
                counter += 1

            newValues = []
            for row in current:
                newRow = list(row)
                newRow.pop(index)
                newValues.append(tuple(newRow))
            newValues = tuple(newValues)

            #print newFieldNames
            #print newValues
            
            self.DROP(table)
            conn.commit()
            self.CREATE(table, fields)
            conn.commit()
            
            string_ = ''
            for i in newValues[0]:
                string_ += ', ?'
            sql = 'INSERT INTO '+table+' ('+', '.join(newFieldNames)+') VALUES ('+string_[2:]+')'
            for values in newValues:
                self.execute(sql,values)
            conn.commit()
        else:
            raise lemon.db.DBError("The 'mode' attribute can only be 'add' or 'drop'.")
            
    def function(self, func, field, table, where=None):
        """Warning: Please note.. no conversion is done here."""
        if func.upper() in ['MAX', 'MIN', 'SUM', 'AVG']:
            self.table = table
            self.fields = field
            #result = "SELECT %s(%s) AS %s FROM %s" % (func, field, field, table)
            result = "SELECT * FROM %s" % (table)
            if where:
                result += " WHERE %s" % where
            self.execute(result)
            if self.fetchall():
                result = "SELECT %s(%s) FROM %s" % (func, field, table)
                if where:
                    result += " WHERE %s" % where
                    
                if self.autoExecute is True:
                    self.execute(result)
                    val = self.fetchall()[0][0]
                    return val
                else:
                    return result
            else:
                return None
        else:
            raise lemon.db.DBError("The function '%s' is not supported by the lemon.db module."%func)
            